"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.completion = exports.commandNames = exports.shorthands = void 0;
exports.rcOptionsTypes = rcOptionsTypes;
exports.cliOptionsTypes = cliOptionsTypes;
exports.help = help;
exports.handler = handler;
const cli_utils_1 = require("@pnpm/cli-utils");
const common_cli_options_help_1 = require("@pnpm/common-cli-options-help");
const config_1 = require("@pnpm/config");
const logger_1 = require("@pnpm/logger");
const matcher_1 = require("@pnpm/matcher");
const outdated_1 = require("@pnpm/outdated");
const error_1 = require("@pnpm/error");
const plugin_commands_env_1 = require("@pnpm/plugin-commands-env");
const enquirer_1 = require("enquirer");
const chalk_1 = __importDefault(require("chalk"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const pluck_1 = __importDefault(require("ramda/src/pluck"));
const unnest_1 = __importDefault(require("ramda/src/unnest"));
const render_help_1 = __importDefault(require("render-help"));
const installDeps_1 = require("../installDeps");
const getUpdateChoices_1 = require("./getUpdateChoices");
const recursive_1 = require("../recursive");
function rcOptionsTypes() {
    return (0, pick_1.default)([
        'cache-dir',
        'dangerously-allow-all-builds',
        'depth',
        'dev',
        'engine-strict',
        'fetch-retries',
        'fetch-retry-factor',
        'fetch-retry-maxtimeout',
        'fetch-retry-mintimeout',
        'fetch-timeout',
        'force',
        'global-dir',
        'global-pnpmfile',
        'global',
        'https-proxy',
        'ignore-pnpmfile',
        'ignore-scripts',
        'lockfile-dir',
        'lockfile-directory',
        'lockfile-only',
        'lockfile',
        'lockfile-include-tarball-url',
        'network-concurrency',
        'noproxy',
        'npm-path',
        'offline',
        'only',
        'optional',
        'package-import-method',
        'pnpmfile',
        'prefer-offline',
        'production',
        'proxy',
        'registry',
        'reporter',
        'save',
        'save-exact',
        'save-prefix',
        'save-workspace-protocol',
        'scripts-prepend-node-path',
        'shamefully-flatten',
        'shamefully-hoist',
        'shared-workspace-lockfile',
        'side-effects-cache-readonly',
        'side-effects-cache',
        'store-dir',
        'unsafe-perm',
        'use-running-store-server',
    ], config_1.types);
}
function cliOptionsTypes() {
    return {
        ...rcOptionsTypes(),
        interactive: Boolean,
        latest: Boolean,
        recursive: Boolean,
        workspace: Boolean,
    };
}
exports.shorthands = {
    D: '--dev',
    P: '--production',
};
exports.commandNames = ['update', 'up', 'upgrade'];
const completion = async (cliOpts) => {
    return (0, cli_utils_1.readDepNameCompletions)(cliOpts.dir);
};
exports.completion = completion;
function help() {
    return (0, render_help_1.default)({
        aliases: ['up', 'upgrade'],
        description: 'Updates packages to their latest version based on the specified range. You can use "*" in package name to update all packages with the same pattern.',
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    {
                        description: 'Update in every package found in subdirectories \
or every workspace package, when executed inside a workspace. \
For options that may be used with `-r`, see "pnpm help recursive"',
                        name: '--recursive',
                        shortAlias: '-r',
                    },
                    {
                        description: 'Update globally installed packages',
                        name: '--global',
                        shortAlias: '-g',
                    },
                    {
                        description: 'How deep should levels of dependencies be inspected. Infinity is default. 0 would mean top-level dependencies only',
                        name: '--depth <number>',
                    },
                    {
                        description: 'Ignore version ranges in package.json',
                        name: '--latest',
                        shortAlias: '-L',
                    },
                    {
                        description: 'Update packages only in "dependencies" and "optionalDependencies"',
                        name: '--prod',
                        shortAlias: '-P',
                    },
                    {
                        description: 'Update packages only in "devDependencies"',
                        name: '--dev',
                        shortAlias: '-D',
                    },
                    {
                        description: 'Don\'t update packages in "optionalDependencies"',
                        name: '--no-optional',
                    },
                    {
                        description: 'Tries to link all packages from the workspace. \
Versions are updated to match the versions of packages inside the workspace. \
If specific packages are updated, the command will fail if any of the updated \
dependencies is not found inside the workspace',
                        name: '--workspace',
                    },
                    {
                        description: 'Show outdated dependencies and select which ones to update',
                        name: '--interactive',
                        shortAlias: '-i',
                    },
                    common_cli_options_help_1.OPTIONS.globalDir,
                    ...common_cli_options_help_1.UNIVERSAL_OPTIONS,
                ],
            },
            common_cli_options_help_1.FILTERING,
        ],
        url: (0, cli_utils_1.docsUrl)('update'),
        usages: ['pnpm update [-g] [<pkg>...]'],
    });
}
async function handler(opts, params = []) {
    if (opts.global && opts.rootProjectManifest == null) {
        return 'No global packages found';
    }
    if (opts.interactive) {
        return interactiveUpdate(params, opts);
    }
    return update(params, opts);
}
async function interactiveUpdate(input, opts) {
    const include = makeIncludeDependenciesFromCLI(opts.cliOptions);
    const projects = (opts.selectedProjectsGraph != null)
        ? Object.values(opts.selectedProjectsGraph).map((wsPkg) => wsPkg.package)
        : [
            {
                rootDir: opts.dir,
                manifest: await (0, cli_utils_1.readProjectManifestOnly)(opts.dir, opts),
            },
        ];
    const rootDir = opts.workspaceDir ?? opts.dir;
    const rootProject = projects.find((project) => project.rootDir === rootDir);
    const outdatedPkgsOfProjects = await (0, outdated_1.outdatedDepsOfProjects)(projects, input, {
        ...opts,
        compatible: opts.latest !== true,
        ignoreDependencies: rootProject?.manifest?.pnpm?.updateConfig?.ignoreDependencies,
        include,
        retry: {
            factor: opts.fetchRetryFactor,
            maxTimeout: opts.fetchRetryMaxtimeout,
            minTimeout: opts.fetchRetryMintimeout,
            retries: opts.fetchRetries,
        },
        timeout: opts.fetchTimeout,
    });
    const workspacesEnabled = !!opts.workspaceDir;
    const choices = (0, getUpdateChoices_1.getUpdateChoices)((0, unnest_1.default)(outdatedPkgsOfProjects), workspacesEnabled);
    if (choices.length === 0) {
        if (opts.latest) {
            return 'All of your dependencies are already up to date';
        }
        return 'All of your dependencies are already up to date inside the specified ranges. Use the --latest option to update the ranges in package.json';
    }
    const { updateDependencies } = await (0, enquirer_1.prompt)({
        choices,
        footer: '\nEnter to start updating. Ctrl-c to cancel.',
        indicator(state, choice) {
            return ` ${choice.enabled ? '●' : '○'}`;
        },
        message: 'Choose which packages to update ' +
            `(Press ${chalk_1.default.cyan('<space>')} to select, ` +
            `${chalk_1.default.cyan('<a>')} to toggle all, ` +
            `${chalk_1.default.cyan('<i>')} to invert selection)`,
        name: 'updateDependencies',
        pointer: '❯',
        result() {
            return this.selected;
        },
        format() {
            if (!this.state.submitted || this.state.cancelled)
                return '';
            if (Array.isArray(this.selected)) {
                return this.selected
                    // The custom format function is used to filter out "[dependencies]" or "[devDependencies]" from the output.
                    // https://github.com/enquirer/enquirer/blob/master/lib/prompts/select.js#L98
                    .filter((choice) => !/^\[.+\]$/.test(choice.name))
                    .map((choice) => this.styles.primary(choice.name)).join(', ');
            }
            return this.styles.primary(this.selected.name);
        },
        styles: {
            dark: chalk_1.default.reset,
            em: chalk_1.default.bgBlack.whiteBright,
            success: chalk_1.default.reset,
        },
        type: 'multiselect',
        validate(value) {
            if (value.length === 0) {
                return 'You must choose at least one package.';
            }
            return true;
        },
        // For Vim users (related: https://github.com/enquirer/enquirer/pull/163)
        j() {
            return this.down();
        },
        k() {
            return this.up();
        },
        cancel() {
            // By default, canceling the prompt via Ctrl+c throws an empty string.
            // The custom cancel function prevents that behavior.
            // Otherwise, pnpm CLI would print an error and confuse users.
            // See related issue: https://github.com/enquirer/enquirer/issues/225
            (0, logger_1.globalInfo)('Update canceled');
            process.exit(0);
        },
    }); // eslint-disable-line @typescript-eslint/no-explicit-any
    const updatePkgNames = (0, pluck_1.default)('value', updateDependencies);
    return update(updatePkgNames, opts);
}
async function update(dependencies, opts) {
    if (opts.latest) {
        const dependenciesWithTags = dependencies.filter((name) => (0, recursive_1.parseUpdateParam)(name).versionSpec != null);
        if (dependenciesWithTags.length) {
            throw new error_1.PnpmError('LATEST_WITH_SPEC', `Specs are not allowed to be used with --latest (${dependenciesWithTags.join(', ')})`);
        }
    }
    const includeDirect = makeIncludeDependenciesFromCLI(opts.cliOptions);
    const include = {
        dependencies: opts.rawConfig.production !== false,
        devDependencies: opts.rawConfig.dev !== false,
        optionalDependencies: opts.rawConfig.optional !== false,
    };
    const depth = opts.depth ?? Infinity;
    return (0, installDeps_1.installDeps)({
        ...opts,
        allowNew: false,
        depth,
        ignoreCurrentSpecifiers: false,
        includeDirect,
        include,
        update: true,
        updateToLatest: opts.latest,
        updateMatching: (dependencies.length > 0) && dependencies.every(dep => !dep.substring(1).includes('@')) && depth > 0 && !opts.latest
            ? (0, matcher_1.createMatcher)(dependencies)
            : undefined,
        updatePackageManifest: opts.save !== false,
        resolutionMode: opts.save === false ? 'highest' : opts.resolutionMode,
        prepareExecutionEnv: plugin_commands_env_1.prepareExecutionEnv.bind(null, opts),
    }, dependencies);
}
function makeIncludeDependenciesFromCLI(opts) {
    return {
        dependencies: opts.production === true || (opts.dev !== true && opts.optional !== true),
        devDependencies: opts.dev === true || (opts.production !== true && opts.optional !== true),
        optionalDependencies: opts.optional === true || (opts.production !== true && opts.dev !== true),
    };
}
//# sourceMappingURL=index.js.map